"""
Created on 2024-10-07

@author: wf
"""
import re
import io
from dataclasses import dataclass
from datetime import date
from typing import Dict, List, Optional, Tuple

from beancount import loader
from beancount.core import amount, data
from beancount.parser import printer

from nomina.stats import Stats


@dataclass
class Preamble:
    start_date: str
    end_date: str
    title: str
    currency: Optional[str] = "EUR"

    def as_header(self) -> str:
        preamble = f"""\
;; -*- mode: org; mode: beancount; -*-
;; Generated by pynomina.beancount
;; Dates: {self.start_date} - {self.end_date}

* Options

option "title" "{self.title}"
option "operating_currency" "{self.currency}"

"""
        return preamble


class Beancount:
    """
    Encapsulates Beancount-specific operations and conversions
    """

    def __init__(self):
        """
        constructor
        """
        self.entries: List[data.Directive] = []
        self.errors: List[str] = []
        self.options_map: Dict = {}

    def load_file(self, file_path: str) -> None:
        """
        Load a Beancount file

        Args:
            file_path (str): Path to the Beancount file
        """
        self.entries, self.errors, self.options_map = loader.load_file(file_path)

    def load_string(self, beancount_string: str) -> None:
        """
        Load Beancount data from a string

        Args:
            beancount_string (str): Beancount data as a string
        """
        self.entries, self.errors, self.options_map = loader.load_string(
            beancount_string
        )

    def add_entry(self, entry: data.Directive):
        """
        add the given entry to my entries
        """
        if entry:
            self.entries.append(entry)

    def sanitize_account_name(self, fq_name) -> str:
        """
        Sanitize the account name by replacing spaces, slashes, dots, commas,
        and removing paired square brackets but keeping the content inside.

        https://beancount.github.io/docs/beancount_language_syntax.html#syntax-overview
        """
        # Remove square brackets but keep the content inside using a named group
        fq_name = re.sub(r"\[(?P<content>.*?)\]", r"\g<content>", fq_name)
        # Replace spaces, slashes, dots, and commas with hyphens
        fixed_name = re.sub(r"[ /\.,%]", "-", fq_name)
        return fixed_name



    def create_open_directive(
        self,
        fq_account_name: str,
        date: date,
        currencies: Optional[List[str]] = None,
        meta: Optional[Dict] = None,
    ) -> data.Open:
        """
        Create a Beancount Open account directive

        Args:
            fq_account_name (str): fully qualifying Account name
            date (date): Opening date
            currencies (Optional[List[str]]): List of currencies for the account
            meta (Optional[Dict]): Metadata for the Open directive

        Returns:
            data.Open: Beancount Open directive
        """
        od = data.Open(meta or {}, date, fq_account_name, currencies or [], None)
        return od

    def create_transaction(
        self,
        date: date,
        description: str,
        postings: List[Tuple[str, float, str]],
        payee: str = None,
        metadata: Dict = None,
    ) -> data.Transaction:
        """
        Create a Beancount Transaction directive

        Args:
            date (date): Transaction date
            description (str): Transaction description
            postings (List[Tuple[str, float, str]]): List of (account, amount, currency) tuples
            payee (str, optional): Payee name
            metadata (Dict, optional): Transaction metadata

        Returns:
            data.Transaction: Beancount Transaction directive
        """
        meta = data.new_metadata("", 0, metadata)
        txn_postings = []
        for account, amount_value, currency in postings:
            amt = amount.Amount(amount.D(str(amount_value)), currency)
            txn_postings.append(data.Posting(account, amt, None, None, None, None))

        tx = data.Transaction(
            meta,
            date,
            "*",
            payee,
            description,
            data.EMPTY_SET,
            data.EMPTY_SET,
            txn_postings,
        )
        return tx

    def entries_to_string(
        self, preamble: Preamble, entries: Optional[List[data.Directive]] = None
    ) -> str:
        """
        Get a text version of the entries, including the preamble.

        Args:
            preamble (Preamble): The preamble object to generate the header.
            entries (Optional[List[data.Directive]]): List of Beancount directives. If None, use self.entries.

        Returns:
            str: The Beancount file content as a string.
        """
        entries_to_print = entries if entries is not None else self.entries
        output = io.StringIO()

        # Add the preamble header
        if preamble:
            output.write(preamble.as_header())

        output.write("* Expenses\n")
        # Print entries to the output
        printer.print_entries(entries_to_print, file=output)

        return output.getvalue()

    def get_stats(self) -> Stats:
        """
        Get statistics about the Beancount data

        Returns:
            Stats: An object containing various statistics about the Beancount data
        """
        accounts = set()
        transactions = 0
        dates = []

        for entry in self.entries:
            if isinstance(entry, data.Open):
                accounts.add(entry.account)
            elif isinstance(entry, data.Transaction):
                transactions += 1
                dates.append(entry.date)

        start_date = min(dates) if dates else None
        end_date = max(dates) if dates else None

        return Stats(
            accounts=len(accounts),
            transactions=transactions,
            start_date=start_date,
            end_date=end_date,
            errors=len(self.errors),
            other={
                "commodities": len(
                    set(
                        entry.currency
                        for entry in self.entries
                        if isinstance(entry, data.Commodity)
                    )
                ),
            },
        )



"""
Created on 2024-10-05

@author: wf
"""

from pathlib import Path
from typing import List, TextIO

from nomina.bzv import Account
from nomina.bzv import Book as BzvBook
from nomina.bzv import Transaction as BzvTransaction
from nomina.file_formats import AccountingFileFormats
from nomina.ledger import Account as LedgerAccount
from nomina.ledger import Book as LedgerBook
from nomina.ledger import Split as LedgerSplit
from nomina.ledger import Transaction as LedgerTransaction
from nomina.nomina_converter import BaseFromLedgerConverter, BaseToLedgerConverter


class BankingZVToLedgerConverter(BaseToLedgerConverter):
    """
    convert BankingZV Entries to Ledger format
    """

    def __init__(self, debug: bool = False):
        """
        Constructor for BankingZV to Ledger Book conversion.

        Args:
            debug (bool): Whether to enable debug logging.
        """
        super().__init__(from_format_acronym="BZV-YAML", debug=debug)

    def load(self, input_path: Path)->BzvBook:
        bzv_book=BzvBook.load_from_file(input_path)
        self.set_source(bzv_book)
        return bzv_book

    def set_source(self,bzv_book:BzvBook):
        self.bzv_book = bzv_book
        self.source = self.bzv_book
        return self.source

    def create_ledger_account(self, bzv_account: Account) -> LedgerAccount:
        """
        create a ledger account from a banking ZV account
        """
        account = LedgerAccount(
            account_id=bzv_account.account_id,
            name=bzv_account.name,
            account_type="BANK" if ":" not in bzv_account.account_id else "CATEGORY",
            description=f"",
            currency="EUR",
            parent_account_id=bzv_account.parent_account_id,
        )
        return account

    def create_ledger_splits(self, transaction: BzvTransaction) -> List[LedgerSplit]:
        """
        create the ledger split
        """
        amount = float(transaction.Amt)
        if transaction.CdtDbtInd == "DBIT":
            amount = -amount

        splits = [
            LedgerSplit(
                amount=amount,
                account_id=transaction.AcctId,
                memo=transaction.RmtInf or "",
            )
        ]

        if transaction.Category:
            splits.append(
                LedgerSplit(
                    amount=-amount,
                    account_id=transaction.Category,
                    memo=transaction.RmtInf or "",
                )
            )

        return splits

    def create_ledger_transaction(
        self, transaction: BzvTransaction
    ) -> LedgerTransaction:
        return LedgerTransaction(
            isodate=transaction.BookgDt,
            description=transaction.BookgTxt or "No description",
            splits=self.create_ledger_splits(transaction),
            memo=transaction.RmtInf or "",
        )

    def convert_to_target(self) -> LedgerBook:
        ledger_book = LedgerBook(
            owner=self.bzv_book.owner, url=self.bzv_book.url, name=self.bzv_book.name
        )

        for bzv_account in self.bzv_book.accounts.values():
            ledger_account = self.create_ledger_account(bzv_account)
            ledger_book.accounts[ledger_account.account_id] = ledger_account

        for transaction in self.bzv_book.transactions:
            ledger_transaction = self.create_ledger_transaction(transaction)
            transaction_id = f"{ledger_transaction.isodate}:{transaction.Id}"
            ledger_book.transactions[transaction_id] = ledger_transaction
        self.target = ledger_book
        return ledger_book

    def to_text(self) -> str:
        yaml_str = self.target.to_yaml()
        return yaml_str



"""
Created on 2024-10-06

@author: wf
"""

import os

from nomina.gnc_ledger import GnuCashToLedgerConverter, LedgerToGnuCashConverter
from tests.basetest import Basetest
from tests.example_testcases import NominaExample


class Test_LedgerGnuCash(Basetest):
    """
    test GnuCash handling
    """

    def setUp(self, debug=True, profile=True):
        Basetest.setUp(self, debug=debug, profile=profile)
        self.examples = NominaExample.get_examples()

    def test_ledger2gnucash(self):
        """
        test converting ledger format to GnuCash
        """
        for name, example in self.examples.items():
            with self.subTest(f"Testing {name}"):
                l2g = LedgerToGnuCashConverter(debug=self.debug)
                output_path = os.path.join("/tmp", f"{name}_l2g_xml.gnucash")
                ledger_book = l2g.load(example.ledger_file)
                _target_object = l2g.convert_from_ledger(ledger_book)
                with open(output_path, "w") as gc_file:
                    output_text = l2g.to_text()
                    gc_file.write(output_text)
                if self.debug:
                    l2g.show_stats()

    def test_gnucash2ledger(self):
        """
        test converting gnu cash to ledger format
        """
        for name, example in self.examples.items():
            with self.subTest(f"Testing {name}"):
                g2l = GnuCashToLedgerConverter(debug=self.debug)
                output_path = os.path.join("/tmp", f"{name}_g2l.yaml")
                ledger_book = g2l.convert_to_ledger(example.gnu_cash_xml_file)
                with open(output_path, "w") as ledger_file:
                    ledger_book.to_yaml(ledger_file)
                if self.debug:
                    g2l.show_stats()



"""
Created on 2024-10-09

@author: wf
"""

from pathlib import Path

from nomina.msmoney_ledger import MicrosoftMoneyToLedgerConverter
from tests.basetest import Basetest


class Test_LedgerMicrosoftMoney(Basetest):
    """
    test Microsoft Money handling
    """

    def setUp(self, debug=False, profile=True):
        Basetest.setUp(self, debug=debug, profile=profile)

    def test_msmoney2ledger(self):
        """
        test the Microsoft Money example
        """
        msmoney_example = Path(self.examples_path) / "sample_microsoft_money.zip"
        converter = MicrosoftMoneyToLedgerConverter(debug=self.debug)
        converter.load(msmoney_example)
        converter.log.do_print = self.debug
        book = converter.convert_to_target()
        converter.show_stats()



"""
Created on 09.10.2024

@author: wf
"""

import json

import networkx as nx


class Graph:
    """
    graph handling library
    """

    def __init__(self):
        self.graph = nx.Graph()

    def add_nodes_from_json(self, file_name, file_content):
        """
        Add nodes from JSON data to the graph
        """
        table_name = file_name.split(".")[0]  # Remove the .json extension

        for line in file_content:
            line = line.strip()
            if line:  # Skip empty lines
                try:
                    json_data = json.loads(line)
                    self.add_node(table_name, json_data)
                except json.JSONDecodeError as e:
                    print(f"Error decoding JSON object in file {file_name}: {e}")

    def add_node(self, table_name, data):
        """
        Add a single node to the graph
        """
        node_id = data.get("id", str(len(self.graph)))
        self.graph.add_node(node_id, type=table_name, **data)

    def dump(self, node_types=None, limit: int = 10):
        """
        Dump the content of the graph for investigation.

        Args:
            node_types (list): List of node types to dump. If None, dump all types.
            limit (int): Maximum number of nodes to dump for each node type. Default is 10.
        """
        self._print_graph_summary()
        node_types = self._get_node_types_to_dump(node_types)
        self._dump_nodes(node_types, limit)
        self._dump_edges(limit)

    def _print_graph_summary(self):
        print(f"Total nodes in graph: {len(self.graph.nodes)}")
        print(f"Total edges in graph: {len(self.graph.edges)}")

        all_node_types = set(
            data.get("type", "Unknown") for _, data in self.graph.nodes(data=True)
        )
        print(f"All node types: {all_node_types}")

    def _get_node_types_to_dump(self, requested_types):
        all_node_types = set(
            data.get("type", "Unknown") for _, data in self.graph.nodes(data=True)
        )
        if requested_types is None:
            return all_node_types
        return (
            set(requested_types) & all_node_types
        )  # Ensure we only dump existing types

    def _dump_nodes(self, node_types, limit):
        print(f"Dumping node types: {node_types}")
        for node_type in node_types:
            self._dump_nodes_of_type(node_type, limit)

    def _dump_nodes_of_type(self, node_type, limit):
        print(f"\nDumping up to {limit} nodes of type '{node_type}':")
        count = 0
        for node, data in self.graph.nodes(data=True):
            if data.get("type") == node_type:
                self._print_node(node, data)
                count += 1
                if count >= limit:
                    break
        if count >= limit:
            remaining = (
                sum(
                    1
                    for _, d in self.graph.nodes(data=True)
                    if d.get("type") == node_type
                )
                - limit
            )
            print(f"  ... and {remaining} more")

    def _print_node(self, node, data):
        print(f"  Node: {node}")
        for key, value in data.items():
            if key != "type":
                print(f"    {key}: {value}")

    def _dump_edges(self, limit):
        if self.graph.edges:
            print("\nSample of edges:")
            for i, (u, v, data) in enumerate(self.graph.edges(data=True)):
                print(f"  Edge {i}: {u} -> {v}")
                for key, value in data.items():
                    print(f"    {key}: {value}")
                if i >= limit - 1:
                    remaining = len(self.graph.edges) - limit
                    if remaining > 0:
                        print(f"  ... and {remaining} more")
                    break
        else:
            print("\nNo edges in the graph.")



"""
Banking ZV Subsembly json export
Created on 2024-10-05

@author: wf
"""

import json
import os
from dataclasses import dataclass, field
from datetime import datetime
from typing import Dict, List, Optional

from dacite import from_dict
from lodstorage.yamlable import lod_storable

from nomina.stats import Stats


@lod_storable
class Transaction:
    Id: str
    AcctId: str
    OwnrAcctCcy: str
    OwnrAcctIBAN: Optional[str]
    OwnrAcctNo: str
    OwnrAcctBIC: Optional[str]
    OwnrAcctBankCode: str
    BookgDt: str
    ValDt: Optional[str]
    Amt: str
    AmtCcy: str
    CdtDbtInd: str
    RmtInf: Optional[str]
    BookgTxt: Optional[str]
    PrimaNotaNo: Optional[str]
    BankRef: Optional[str]
    BkTxCd: Optional[str]
    BookgSts: str
    GVC: Optional[str]
    Category: Optional[str]
    ReadStatus: bool
    Flag: str


@lod_storable
class Account:
    account_id: str
    name: str
    parent_account_id: Optional[str] = None


@lod_storable
class Book:
    """
    a Banking ZV Book
    """

    name: str
    owner: Optional[str] = (None,)
    url: Optional[str] = (None,)
    since: Optional[str] = (None,)
    account_json_exports: Dict[str, str] = field(default_factory=dict)

    def __post_init__(self):
        self.accounts: Dict[str, Account] = {}
        self.transactions: List[Transaction] = []

    @classmethod
    def load_from_file(cls, filename: str) -> "Book":
        book = Book.load_from_yaml_file(filename)
        yaml_dir = os.path.dirname(filename)

        # Adjust JSON file paths to be relative to the YAML file
        for account, json_path in book.account_json_exports.items():
            book.account_json_exports[account] = os.path.join(yaml_dir, json_path)

        book.load_accounts_from_accounts_dict(book.account_json_exports)
        return book

    def get_stats(self) -> Stats:
        """
        Calculates and returns statistics related to the transactions, accounts, and other properties of the Book.

        Returns:
            Stats: An object containing various statistics about the Book.
        """
        dates = [
            datetime.strptime(tx.BookgDt, "%Y-%m-%d")
            for tx in self.transactions
            if tx.BookgDt
        ]

        currencies = {}
        categories = set()
        for tx in self.transactions:
            currencies[tx.AmtCcy] = currencies.get(tx.AmtCcy, 0) + 1
            if tx.Category:
                categories.add(tx.Category)

        return Stats(
            accounts=len(self.accounts),
            transactions=len(self.transactions),
            start_date=min(dates).strftime("%Y-%m-%d") if dates else None,
            end_date=max(dates).strftime("%Y-%m-%d") if dates else None,
            categories=len(categories),
            currencies=currencies,
            other={"name": self.name, "owner": self.owner},
        )

    def load_accounts_from_accounts_dict(self, account_json_dict: dict):
        """
        Loads transactions from a dictionary of JSON file paths and populates the accounts.

        Args:
            account_json_dict (dict): A dictionary where keys are account IDs and values are paths to JSON files containing transaction data.

        """
        for bank_account, json_file in account_json_dict.items():
            # Load transactions from the JSON file
            txs = self.load_transactions_from_json_file(json_file)
            self.transactions.extend(txs)
            # get the account id from the first transaction
            account_id = txs[0].AcctId

            # Create an account object and add it to the accounts dictionary
            self.accounts[account_id] = Account(
                account_id=account_id, name=bank_account
            )

        # Create category accounts as needed
        self.create_category_accounts()

    def add_category_account(self, category: str):
        """
        add a category account for the given category
        """
        parts = category.split(":")
        current = ""
        parent = None
        for part in parts:
            if current:
                parent = current
                current += ":"
            current += part
            if current not in self.accounts:
                account = Account(
                    account_id=current, name=part, parent_account_id=parent
                )
                self.accounts[current] = account

    def create_category_accounts(self):
        for transaction in self.transactions:
            category = transaction.Category
            if category:
                self.add_category_account(category)

    @staticmethod
    def load_transactions_from_json_file(json_file_path: str) -> List[Transaction]:
        with open(json_file_path, "r", encoding="utf-8") as json_file:
            trecs = json.load(json_file)
            return [from_dict(Transaction, trec) for trec in trecs]



"""
Created on 09.10.2024

@author: wf
"""

from pathlib import Path
from nomina.date_utils import DateUtils
from nomina.ledger import Account, Book, Split, Transaction
from nomina.msmoney import MsMoney
from nomina.nomina_converter import BaseToLedgerConverter


class MicrosoftMoneyToLedgerConverter(BaseToLedgerConverter):
    """
    Microsoft Money to Ledger Converter
    """

    def __init__(self, debug: bool = False):
        """
        Constructor for Microsoft Money to Ledger Book conversion.

        Args:
            debug (bool): Whether to enable debug logging.
        """
        super().__init__(from_format_acronym="MONEY", debug=debug)
        self.ms_money = None

    def load(self, input_path: Path) -> MsMoney:
        """
        Load Microsoft Money data
        """
        self.ms_money = MsMoney()
        self.ms_money.load(str(input_path))
        self.source = self.ms_money
        return self.source

    def convert_to_target(self) -> Book:
        """
        convert the microsoft money entries to a Ledger Book
        """
        book = Book()
        book.name = self.ms_money.header.name if self.ms_money.header else "Unknown"
        book.since = self.ms_money.header.date if self.ms_money.header else None
        graph=self.ms_money.graph.graph
        self.log.log("✅", "graph", f"Total nodes: {len(graph.nodes)}")
        node_types = set(
            data.get("type", "Unknown")
            for _, data in graph.nodes(data=True)
        )
        self.log.log("✅", "graph", f"Node types: {node_types}")

        # Create accounts
        for node, data in graph.nodes(data=True):
            if data.get("type") == "ACCT":
                account = Account(
                    account_id=str(data.get('hacct')),
                    name=data.get("szFull", ""),
                    account_type=data.get("acct_type", "EXPENSE"),
                    description=data.get("desc", ""),
                    currency=data.get("currency", "EUR"),
                )
                book.add_account(account)

        self.log.log("✅", "accounts", f"Accounts created: {len(book.accounts)}")

        # Create transactions
        for _node, data in graph.nodes(data=True):
            tx_dict=self.ms_money.to_transaction_dict(data)
            if tx_dict is not None:
                isodate=tx_dict["isodate"]
                amount=tx_dict["amount"]
                transaction_id=tx_dict["transaction_id"]
                account_id=tx_dict["account_id"]
                transaction = Transaction(
                    isodate=isodate,
                    description=f"Transaction {transaction_id}",  # No clear description field, using a placeholder
                    splits=[],  # We'll handle splits later
                    payee="",  # No clear payee field in the example
                    memo=f"Amount: {amount}"
                )

                # Add a single split for now, we'll refine this later
                split = Split(
                    amount=float(amount),
                    account_id=account_id,
                    memo=f"Transaction {transaction_id}"
                )
                transaction.splits.append(split)

                book.transactions[transaction_id] = transaction

        self.log.log(
            "✅", "transactions", f"Transactions created: {len(book.transactions)}"
        )

        self.target = book
        return book

    def to_text(self) -> str:
        yaml_str = self.target.to_yaml()
        return yaml_str



"""
Created on 2024-10-06

@author: wf
"""

import os
import re
import zipfile
from dataclasses import dataclass
from typing import Dict, List, Optional

import chardet


@dataclass
class AccountingFileFormat:
    """
    A personal account file format
    """

    name: str
    acronym: str
    ext: str
    wikidata_id: str
    content_pattern: str
    pattern_file: Optional[str] = None
    encoding: str = "utf-8"


class AccountingFileFormats:
    """
    Detector for various accounting file formats
    """

    def __init__(self, debug: bool = False):
        self.debug = debug
        self.formats: List[AccountingFileFormat] = [
            AccountingFileFormat(
                name="Beancount",
                acronym="BEAN",
                ext=".beancount",
                wikidata_id="Q130456404",
                content_pattern=r'option "title"',
            ),
            AccountingFileFormat(
                name="GnuCash XML",
                acronym="GC-XML",
                ext=".gnucash",
                wikidata_id="Q130445392",
                content_pattern=r"<gnc-v2",
            ),
            AccountingFileFormat(
                name="GnuCash SQLite",
                acronym="GC-SQLITE",
                ext=".gnucash",
                wikidata_id="Q130445392",
                content_pattern=r"SQLite format 3",
            ),
            AccountingFileFormat(
                name="Microsoft Money",
                acronym="MONEY",
                ext=".zip",
                wikidata_id="Q117428",
                pattern_file="nomina.yaml",
                content_pattern=r"NOMINA-MICROSOFT-MONEY-YAML",
                encoding="utf-8",
            ),
            AccountingFileFormat(
                name="pyNomina Ledger Book YAML",
                acronym="LB-YAML",
                ext=".yaml",
                wikidata_id="Q281876",
                content_pattern=r"file_type:\s*NOMINA-LEDGER-BOOK-YAML|accounts:\s*\w+:",
            ),
            AccountingFileFormat(
                name="FinanzmanagerDeluxe",
                acronym="FMD",
                ext=".qif",
                wikidata_id="Q750657",
                content_pattern=r"!Option:MDY",
                encoding="iso-8859-1",
            ),
            AccountingFileFormat(
                name="Quicken Interchange Format",
                acronym="QIF",
                ext=".qif",
                wikidata_id="Q750657",
                content_pattern=r"!Account.*!Type:Cash",
            ),
            AccountingFileFormat(
                name="Subsembly JSON",
                acronym="BZV-JSON",
                ext=".json",
                wikidata_id="Q130443951",
                content_pattern=r'"AcctId":\s*"[^"]+".*"OwnrAcctCcy":\s*"[^"]+"',
            ),
            AccountingFileFormat(
                name="pyNomina Banking ZV YAML export",
                acronym="BZV-YAML",
                ext=".yaml",
                wikidata_id="Q281876",
                content_pattern=r"account_json_exports\s*:",
            ),
        ]
        self.format_by_acronym: Dict[str, AccountingFileFormat] = {
            fformat.acronym: fformat for fformat in self.formats
        }

    def get_by_acronym(self, acronym: str) -> "AccountingFileFormat":
        """
        get format by acronym
        """
        return self.format_by_acronym.get(acronym)

    def detect_format(self, file_path: str) -> Optional[AccountingFileFormat]:
        """
        Detect the accounting file format based on content and extension
        """
        _, file_extension = os.path.splitext(file_path)

        if file_extension.lower() == ".zip":
            return self._detect_from_zip(file_path)
        else:
            return self._detect_from_regular_file(file_path)

    def _detect_from_zip(self, zip_path: str) -> Optional[AccountingFileFormat]:
        """
        Detect accounting file format by looking inside the zip file.
        """
        with zipfile.ZipFile(zip_path, "r") as zip_file:
            for fformat in self.formats:
                # Check if the format has a pattern_file defined
                if (
                    fformat.ext == ".zip"
                    and fformat.pattern_file in zip_file.namelist()
                ):
                    content = self._extract_file_content_from_zip(
                        zip_file, fformat.pattern_file
                    )
                    if content and self._match_pattern(
                        content, fformat.content_pattern
                    ):
                        return fformat
        return None

    def _extract_file_content_from_zip(
        self, zip_file: zipfile.ZipFile, pattern_file: str
    ) -> Optional[str]:
        """
        Extract the content of a specific file within a zip and decode it.
        """
        with zip_file.open(pattern_file) as file:
            raw_data = file.read()
            return self._decode_content(raw_data)

    def _detect_from_regular_file(
        self, file_path: str
    ) -> Optional[AccountingFileFormat]:
        """
        Detect accounting file format from a regular non-zip file.
        """
        with open(file_path, "rb") as file:
            raw_data = file.read(10000)  # Read the first 10000 bytes
            content = self._decode_content(raw_data)

        if content:
            for fformat in self.formats:
                if os.path.splitext(file_path)[
                    1
                ].lower() == fformat.ext.lower() and self._match_pattern(
                    content, fformat.content_pattern
                ):
                    return fformat
        return None

    def _decode_content(self, raw_data: bytes) -> Optional[str]:
        """
        Detect the encoding of the raw data and decode it into a string.
        """
        result = chardet.detect(raw_data)
        encoding = result.get("encoding", "utf-8")
        try:
            return raw_data.decode(encoding)
        except UnicodeDecodeError:
            return raw_data.decode("latin-1", errors="ignore")

    def _match_pattern(self, content: str, pattern: str) -> bool:
        """
        Check if the content matches the given pattern.
        """
        return bool(re.search(pattern, content, re.DOTALL | re.IGNORECASE))



'''
Created on 2024-10-10

@author: wf
'''
from tests.basetest import Basetest
from nomina.nomina_beancount import Beancount


class Test_Beancount(Basetest):
    """
    test Beancount handling
    """

    def setUp(self, debug=True, profile=True):
        Basetest.setUp(self, debug=debug, profile=profile)

    def test_sanitze_account_names(self):
        """
        test sanitizing account names
        """
        beancount=Beancount()
        for fq_name,expected_sanitized in [
            ("Assets:[Cash]","Assets:Cash"),
            ("Assets:4240-Gas,-Strom,-Wasser","Assets:4240-Gas--Strom--Wasser"),
            ("4660-Reisekosten-Arbeitnehmer:Auto-0.52","4660-Reisekosten-Arbeitnehmer:Auto-0-52"),
            ("3300-Wareneinkauf-7%","3300-Wareneinkauf-7-")
        ]:
            sanitized=beancount.sanitize_account_name(fq_name)
            self.assertEqual(expected_sanitized,sanitized)



"""
Created on 2024-10-07

@author: wf
"""

import re
from datetime import datetime
from typing import Dict, Optional

from beancount.core import data

from nomina.date_utils import DateUtils
from nomina.file_formats import AccountingFileFormats
from nomina.ledger import Account as LedgerAccount
from nomina.ledger import Book as LedgerBook
from nomina.ledger import Split as LedgerSplit
from nomina.ledger import Transaction as LedgerTransaction
from nomina.nomina_beancount import Beancount, Preamble
from nomina.nomina_converter import BaseFromLedgerConverter, BaseToLedgerConverter


class BeancountToLedgerConverter(BaseToLedgerConverter):
    """
    Convert Beancount format to Ledger Book
    """

    def __init__(self, debug: bool = False):
        """
        Constructor for Beancount to Ledger Book conversion.

        Args:
            debug (bool): Whether to enable debug logging.
        """
        super().__init__(
            from_format_acronym="BEAN", debug=debug
        )  # Look up the formats using AccountingFileFormatDetector

        # Initialize instance variables
        self.beancount = None
        self.beancount_file = None
        self.ledger_book = None
        self.account_map = {}

    def load(self, input_path: str) -> Beancount:
        """

        load the beancount file

        Args:
            input_path (str): The path to the input file.
        """
        self.beancount_file = input_path
        beancount = Beancount()
        beancount.load_file(self.beancount_file)
        self.set_source(beancount)
        return self.beancount

    def set_source(self, source: LedgerBook):
        self.source = source
        self.beancount = source

    def convert_to_target(self) -> LedgerBook:
        """Convert the Beancount file to a Ledger Book."""
        self.ledger_book = LedgerBook()
        self.account_map: Dict[str, LedgerAccount] = {}

        for entry in self.beancount.entries:
            if isinstance(entry, data.Open):
                self.convert_account(entry)
            elif isinstance(entry, data.Transaction):
                self.convert_transaction(entry)

        return self.ledger_book

    def convert_account(self, entry: data.Open):
        """Create a Ledger account from a Beancount Open directive."""
        account_parts = entry.account.split(":")
        account_type = account_parts[0].upper()
        account_name = ":".join(account_parts[1:])

        ledger_account = LedgerAccount(
            account_id=entry.account,
            name=account_name,
            account_type=account_type,
            currency=entry.currencies[0] if entry.currencies else "EUR",
            description=entry.meta.get("description", ""),
        )
        self.ledger_book.add_account(ledger_account)
        self.account_map[entry.account] = ledger_account

    def convert_transaction(self, entry: data.Transaction):
        """Create a Ledger transaction from a Beancount Transaction."""
        splits = []
        for posting in entry.postings:
            amount = posting.units.number
            account_id = posting.account

            split = LedgerSplit(
                amount=float(amount),
                account_id=account_id,
                memo=posting.meta.get("memo", ""),
                reconciled=False,  # Beancount doesn't have a direct equivalent to Ledger's reconciled flag
            )
            splits.append(split)

        ledger_transaction = LedgerTransaction(
            isodate=entry.date.isoformat(),
            description=entry.narration,
            splits=splits,
            payee=entry.payee,
            memo=entry.meta.get("memo", ""),
        )

        # Generate a unique ID for the transaction
        transaction_id = (
            f"{ledger_transaction.isodate}:{hash(ledger_transaction.description)}"
        )
        self.ledger_book.transactions[transaction_id] = ledger_transaction


class LedgerToBeancountConverter(BaseFromLedgerConverter):
    """
    converter for Ledger Book to Beancount
    """

    def __init__(self, debug: bool = False):
        """
        Constructor for Ledger Book to Beancount conversion.

        Args:
            debug (bool): Whether to enable debug logging.
        """
        super().__init__(to_format_acronym="BEAN", debug=debug)

        # Initialize instance variables
        self.lbook = None
        self.lbook_stats = None
        self.start_date = None
        self.beancount = None

    def load(self, input_path: str) -> LedgerBook:
        """
        load the Ledger Book

        Returns:
            LedgerBook: the ledger book
        """
        lbook = LedgerBook.load_from_yaml_file(input_path)
        self.set_source(lbook)
        return lbook

    def set_source(self, source: LedgerBook):
        self.source = source
        self.lbook = source

    def convert_to_target(self) -> Beancount:
        """
        Convert the ledger book to a Beancount object.
        Args:
            lbook (LedgerBook): The ledger book to convert.

        Returns:
            Beancount: The converted Beancount object.
        """
        self.lbook = self.source
        self.lbook_stats = self.lbook.get_stats()
        self.start_date = self.lbook.get_stats().start_date or DateUtils.iso_date(
            datetime.now()
        )
        self.beancount = Beancount()

        for account in self.lbook.accounts.values():
            self.beancount.add_entry(self.convert_account(account))

        for transaction in self.lbook.transactions.values():
            self.beancount.add_entry(self.convert_transaction(transaction))

        self.target = self.beancount
        return self.beancount

    def to_text(self):
        beancount = self.target
        preamble = Preamble(
            start_date=self.start_date,
            end_date=self.lbook_stats.end_date or "Unknown",
            title=self.lbook.name or "Converted Ledger",
            currency=self.lbook_stats.main_currency(),
        )
        text = beancount.entries_to_string(preamble)
        return text

    def get_beancount_name_for_account(self, account: LedgerAccount) -> str:
        """
        get the beancount name for the ledger account name
        """
        self.account_type_map = {
            "ROOT": "Equity",
            "BANK": "Assets",
            "EXPENSE": "Expenses",
            "INCOME": "Income",
            "LIABILITY": "Liabilities",
            "EQUITY": "Equity",
            "ASSET": "Assets",
        }
        prefix = self.account_type_map.get(account.account_type, "Assets")
        fq_name = self.lbook.fq_account_name(account)
        fq_name = self.beancount.sanitize_account_name(fq_name)
        if fq_name in self.account_type_map.keys():
            return None
        beancount_account_name = f"{prefix}:{fq_name}"
        return beancount_account_name

    def convert_account(self, account: LedgerAccount) -> Optional[data.Open]:
        """
        Convert a ledger account to a Beancount Open directive.

        Args:
            account (LedgerAccount): The ledger account to convert.

        Returns:
            Optional[data.Open]: The Beancount Open directive, or None if conversion fails.
        """
        currencies = [account.currency] if account.currency else []
        beancount_account_name = self.get_beancount_name_for_account(account)
        if not beancount_account_name:
            return None
        od = self.beancount.create_open_directive(
            fq_account_name=beancount_account_name,
            date=self.start_date,
            currencies=currencies,
            meta={"description": account.description} if account.description else None,
        )
        return od

    def convert_transaction(
        self, transaction: LedgerTransaction
    ) -> Optional[data.Transaction]:
        """
        Convert a ledger transaction to a Beancount Transaction directive.

        Args:
            transaction (Transaction): The ledger transaction to convert.

        Returns:
            Optional[data.Transaction]: The Beancount Transaction directive, or None if conversion fails.
        """
        date = DateUtils.parse_date(transaction.isodate)
        if date is None:
            self.log.log(
                "⚠️", "date_parse", f"Unable to parse date: {transaction.isodate}"
            )
            return None

        postings = []
        for split in transaction.splits:
            if split.account_id not in self.lbook.accounts:
                self.log.log("❌","split",f"invalid split account: {split.account_id}")
                continue
            split_account = self.lbook.accounts[split.account_id]
            beancount_account_name = self.get_beancount_name_for_account(split_account)
            postings.append(
                (beancount_account_name, split.amount, split_account.currency or "EUR")
            )

        if not postings:
            self.log.log(
                "⚠️",
                "transaction_skip",
                f"Skipping transaction with no valid postings: {transaction.description}",
            )
            return None

        return self.beancount.create_transaction(
            date=date,
            description=transaction.description,
            postings=postings,
            payee=getattr(transaction, "payee", None),
            metadata={"memo": transaction.memo} if transaction.memo else None,
        )



"""
Created on 2024-10-09

@author: wf
"""

from nomina.msmoney import MsMoney
from tests.basetest import Basetest


class Test_Microsoft_Money(Basetest):
    """
    test Microsoft money handling
    """

    def setUp(self, debug=False, profile=True):
        Basetest.setUp(self, debug=debug, profile=profile)

    def test_load_microsoft_money(self):
        """
        test loading Microsoft Money zip file
        """
        for money_example in [
            self.examples_path + "/sample_microsoft_money.zip",
        ]:

            ms_money = MsMoney()
            ms_money.load(money_example)
            stats = ms_money.get_stats()
            if self.debug:
                stats.show()

            # Assert that the header was loaded correctly
            self.assertIsNotNone(ms_money.header)
            self.assertEqual(ms_money.header.file_type, "NOMINA-MICROSOFT-MONEY-YAML")
            self.assertEqual(ms_money.header.version, "0.1")

            graph = ms_money.graph.graph

            # Assert that the graph was populated with nodes
            self.assertGreater(len(graph.nodes), 0)

            if self.debug:
                print(f"Loaded {len(graph.nodes)} nodes into the graph")
                print(f"File name: {ms_money.header.name}")
                print(f"File date: {ms_money.header.date}")

                print()
                ms_money.graph.dump(["ACCT", "TRN", "TRN_SPLIT"])



"""
Created on 2024-10-07

@author: wf
"""

from pathlib import Path

from nomina.bzv import Book
from nomina.stats import Stats
from tests.basetest import Basetest
from tests.example_testcases import NominaExample


class Test_BZV(Basetest):
    """
    test Banking ZV
    """

    def setUp(self, debug=True, profile=True):
        Basetest.setUp(self, debug=debug, profile=profile)

    def test_read_bzv(self):
        """
        test reading Banking ZV files
        """
        bzv_example = self.examples_path + "/expenses2024_bzv.yaml"
        book = Book.load_from_file(bzv_example)
        stats = book.get_stats()
        if self.debug:
            stats.show()
        example = NominaExample(
            name="expenses2024",
            owner="John Doe",
            url=None,
            example_path=Path(self.examples_path),
            expected_stats=Stats(accounts=2, transactions=2, currencies={"EUR": 2}),
        )
        example.check_stats(stats)



"""
Created on 2024-10-09

@author: wf
"""

import os
from json import JSONDecodeError
from zipfile import ZipFile

from lodstorage.yamlable import lod_storable

from nomina.date_utils import DateUtils
from nomina.graph import Graph
from nomina.stats import Stats
from typing import Any,Dict

@lod_storable
class ZipHeader:
    file_type: str
    version: str
    name: str
    date: str
    size: int
    sha256: str
    jetversion: str


class MsMoney:
    """
    Wrapper for Microsoft Money content
    """

    def __init__(self):
        self.header = None
        self.graph = Graph()

    def load(self, path: str):
        """
        Load Microsoft Money data from a ZIP file or a directory
        """
        if os.path.isdir(path):
            file_generator = self._generate_files_from_directory(path)
        else:
            file_generator = self._generate_files_from_zip(path)

        self._handle_files(file_generator)

    def _generate_files_from_zip(self, zip_path: str):
        """
        Generator that yields file name and file object from a ZIP file
        """
        with ZipFile(zip_path, "r") as zip_file:
            for file_name in zip_file.namelist():
                with zip_file.open(file_name) as file_content:
                    yield file_name, file_content

    def _generate_files_from_directory(self, dir_path: str):
        """
        Generator that yields file name and file object from a directory
        """
        for file_name in os.listdir(dir_path):
            file_path = os.path.join(dir_path, file_name)
            with open(file_path, "r") as file_content:
                yield file_name, file_content

    def _handle_files(self, file_generator):
        """
        Handle the files yielded by the generator (from ZIP or directory)
        """
        for file_name, file_content in file_generator:
            if file_name == "nomina.yaml":
                # Load the YAML file
                self.header = ZipHeader.load_from_yaml_stream(file_content)
            elif file_name.endswith(".json"):
                try:
                    # Load JSON files as networkx nodes
                    self.graph.add_nodes_from_json(file_name, file_content)
                except JSONDecodeError as e:
                    print(f"Error decoding JSON in file {file_name}: {e}")
                    continue

    def to_transaction_dict(self,data)->Dict[str,Any]:
        """
        convert a data node to a transcaction dict with transaction_id and isodate
        """
        tx_dict=None
        if data.get("type") == "TRN":
            transaction_id = str(data.get('htrn'))
            t_date=data.get('dt')
            isodate=DateUtils.parse_date(t_date)
            tx_dict={
                "transaction_id": transaction_id,
                "isodate":isodate,
                "amount": data.get('amt', 0.0),
                "account_id": data.get('hacct')
            }
        return tx_dict

    def get_stats(self) -> Stats:
        """
        Get statistics about the Microsoft Money data.

        Returns:
            Stats: An object containing various statistics about the data.
        """
        graph = self.graph.graph  # Assign for clarity

        # Calculate date range
        dates = []
        transactions=0
        for _node, data in graph.nodes(data=True):
            tx_dict=self.to_transaction_dict(data)
            if tx_dict is not None:
                transactions+=1
                dates.append(tx_dict["isodate"])

        if dates:
            min_date = min(dates)
            max_date = max(dates)
        else:
            min_date = max_date = None

        # Count accounts and calculate currency counts
        accounts = 0
        currency_counts = {}
        for _node, data in graph.nodes(data=True):
            if data.get("type") == "ACCT":
                accounts += 1
                currency = data.get("currency", "UNKNOWN")
                currency_counts[currency] = currency_counts.get(currency, 0) + 1

        return Stats(
            accounts=accounts,
            transactions=transactions,
            start_date=min_date,
            end_date=max_date,
            currencies=currency_counts,
        )



"""
Created on 2024-10-06

@author: wf
"""

from dataclasses import field
from typing import Any, Dict, Optional

from lodstorage.yamlable import lod_storable


@lod_storable
class Stats:
    """
    Ledger statistics
    """

    accounts: int
    transactions: int
    start_date: Optional[str] = None
    end_date: Optional[str] = None
    classes: Optional[int] = None
    categories: Optional[int] = None
    errors: Optional[int] = None
    currencies: Dict[str, int] = field(default_factory=dict)
    other: Optional[Dict[str, Any]] = field(default_factory=dict)

    def show(self):
        print(f"# Accounts: {self.accounts}")
        print(f"# Transactions: {self.transactions}")
        print(f"Date Range: {self.start_date} to {self.end_date}")

        if self.classes is not None:
            print(f"# Classes: {self.classes}")

        if self.categories is not None:
            print(f"# Categories: {self.categories}")

        if self.errors is not None:
            print(f"# Errors: {self.errors}")

        if self.currencies:
            print(
                f"# Currencies: {', '.join(f'{currency}: {count}' for currency, count in self.currencies.items())}"
            )

        if self.other:
            print("Other Details:")
            for key, value in self.other.items():
                print(f"  {key}: {value}")

    def main_currency(self) -> Optional[str]:
        """
        Determine the main currency used in the ledger.

        Returns:
            Optional[str]: The currency with the highest count, or None if no currencies are present.
        """
        if not self.currencies:
            return None
        return max(self.currencies, key=self.currencies.get)
